module Fedi.Routes.Inbox where

import Fedi.Requests
import Fedi.Types
import Fedi.UserDetails
import Fedi.Routes.Helpers
import Fedi.Helpers
import Web.Twain qualified as Twain
import Web.Twain.Types qualified as Twain
import Data.Text qualified as T
import Network.Wai qualified as Wai
import Control.Monad.IO.Class (liftIO)
import Text.ParserCombinators.ReadP qualified as P
import Data.Text.Encoding qualified as T
import Data.Text.Encoding.Base64 qualified as Base64
import Data.ByteString.Lazy qualified as BSL
import Fedi.Crypto

-- * Inbox

matchInbox :: UserDetails -> Twain.PathPattern
matchInbox details =
  fromString ("/" <> details.username <> "/inbox")

handleInbox :: (AnyActivity -> Twain.ResponderM Twain.Response) -> Twain.ResponderM b
handleInbox handle = do
  activity <- checkSignatureAndParseBody
  response <- handle activity
  Twain.send response

-- | Check the signature of the sender and parse the body of the request.
checkSignatureAndParseBody :: Twain.ResponderM AnyActivity
checkSignatureAndParseBody = do
  -- get info
  request <- Twain.request
  body <- liftIO (Wai.strictRequestBody request)
  sigheader <- parseSignature <$> Twain.header "Signature"
  digest <-
    maybe (error "missing header Digest") T.encodeUtf8 <$> Twain.header "Digest"
  (person :: Person) <- liftIO $ sendGet sigheader.keyId
  let personPkid = person.otype.publicKey.pkid
  let personPublicKey = pemToBS person.otype.publicKey.publicKeyPem

  -- check
  liftIO $
    checkSignature personPkid personPublicKey sigheader digest (BSL.toStrict body)

  -- parse the body and return it
  parseJson body

checkSignature
  :: Url -> ByteString -> SignatureHeader -> ByteString -> ByteString -> IO ()
checkSignature personPkid personPublicKey sigheader digest body = do
  -- check
  unless (personPkid == sigheader.keyId) $
    error "public key mismatch with signature."

  unless (verifyPub personPublicKey sigheader.signature body) $
    error "signature verification failed."

  unless (verifyDigest personPublicKey digest body) $
    error "digest verification failed."
  -- todo: check date

data SignatureHeader
  = SignatureHeader
    { -- | Where to get the public key for this actor
      keyId :: Url
    , -- | Which headers have been sent
      headers :: [T.Text]
    , -- | Contains the signature
      signature :: ByteString
    }

data Component
  = KeyId
  | Headers
  | Signature
  deriving Eq

parseSignature :: Maybe T.Text -> SignatureHeader
parseSignature minput =
  let
    input = maybe (error "no signature.") T.unpack minput
  in case P.readP_to_S parser input of
    [(sig, "")] -> sig
    _ -> error "error parsing signature."
  where
    lookup' a b =
      fromMaybe (error "error parsing signature") $ lookup a b
    parser = do
      components <- component `P.sepBy` P.char ','
      pure SignatureHeader
        { keyId = lookup' KeyId components
        , headers = T.split (==' ') . T.pack $ lookup' Headers components
        , signature =
          ( T.encodeUtf8
          . Base64.decodeBase64Lenient
          . T.pack
          . lookup' Signature
          ) components
        }
    component = P.choice
      [ do
        _ <- P.string "keyId="
        url <- P.between (P.char '\"') (P.char '\"') (P.munch1 (/= '\"'))
        pure (KeyId, url)
      , do
        _ <- P.string "headers="
        url <- P.between (P.char '\"') (P.char '\"') (P.munch1 (/= '\"'))
        pure (Headers, url)
      , do
        _ <- P.string "signature="
        url <- P.between (P.char '\"') (P.char '\"') (P.munch1 (/= '\"'))
        pure (Signature, url)
      ]
